@@ -1,54 +1,12 @@
-# import math
 import fsspec
 import tempfile
 import warnings
 
 import pandas as pd
 
-# class TensorIterable(IterableDataset):
-#     def __init__(self, osds, dtype = pt.float32, max_iter = None):
-#         self.iter = iter(osds)
-#         self.dtype = dtype
-#         self.max_iter = max_iter
-
-#     def __iter__(self):        
-#         val = next(self.iter)
-#         while val is not None:
-#             try:
-#                 yield pt.tensor(val.values.astype(float)).to(dtype = self.dtype)
-#                 val = next(self.iter)
-#             except ValueError as err:
-#                 raise ValueError(f"Failed data type conversion, use BaseObjectStorageDataset(header='infer') if the dataset has a header:  {err}")
-
-#     def __len__(self):
-#         if self.max_iter is None:
-#             return None
-#         else:
-#             return self.max_iter
-
-# class TabularIterableDataset(IterableDataset):
-#     def __init__(self, osds, dtype = None):
-#         self.iter = iter(osds)
-#         self.dtype = dtype
-
-#     def __iter__(self):        
-#         val = next(self.iter)
-#         while val is not None:
-#             try:
-#                 numpy_array = val._get_numeric_data().values
-#                 pt_tensor = pt.from_numpy(numpy_array)                
-#                 pt_tensor = pt_tensor.to(dtype = self.dtype) if self.dtype is not None else pt_tensor
-#                 yield pt_tensor
-#                 val = next(self.iter)
-#             except ValueError as err:
-#                 raise ValueError(f"Failed data type conversion, use BaseObjectStorageDataset(header='infer') if the dataset has a header:  {err}")
-
-#     def __len__(self):
-#         return None
-
-def make_metadata_df(glob = None, df = None):    
+def make_metadata_df(glob = None, df = None, header = None):    
     assert glob is not None or df is not None, "Either glob or df must be specified"
-    df = pd.read_csv(glob) if glob else df
+    df = pd.read_csv(f"filecache::{glob}", header) if glob else df
     assert set(df.columns) == set(['_partitionId', 'count']), "The metadata source must include _partitionId and count columns"
     df = df.sort_values(by='_partitionId', ignore_index=True)
     df['end_idx'] = df['count'].cumsum()
@@ -56,13 +14,16 @@ def make_metadata_df(glob = None, df = None):
     df = df[['_partitionId', 'start_idx', 'end_idx']]
     return df
 
-def objs_to_metadata_df(objs):
+def objs_to_metadata_df(objs, header = None):
     rows = []
     with warnings.catch_warnings():
         warnings.simplefilter("ignore")
         for idx, obj in enumerate(objs):
-            df = pd.read_csv(f"filecache::{obj}", header = None)
-            rows.append({"_partitionId": idx, "count": len(df),})
+            try:
+                df = pd.read_csv(f"filecache::{obj}", header = header)
+                rows.append({"_partitionId": idx, "count": len(df),})
+            except pd.errors.EmptyDataError:                
+                rows.append({"_partitionId": idx, "count": 0,})
     return pd.DataFrame(rows)
 
 def df_read_spec(df, row_count, row_start_idx, row_end_idx):
@@ -75,6 +36,10 @@ def df_read_spec(df, row_count, row_start_idx, row_end_idx):
     #if the row_end_idx matches 2 partitions, calling item results in ValueError: can only convert an array of size 1 to a Python scalar
     end_obj_idx_res = df[ (row_end_idx >= df['start_idx']) & (row_end_idx <= df['end_idx']) ]['_partitionId'].max()    
     end_obj_idx = end_obj_idx_res if isinstance(end_obj_idx_res, int) else end_obj_idx_res.item()
+
+    start_obj_idx = df [ df['_partitionId'] == start_obj_idx ].index.values.item()
+    end_obj_idx = df [ df['_partitionId'] == end_obj_idx ].index.values.item()
+
     if row_end_idx > row_start_idx:
         
         df_spec = df.iloc[start_obj_idx : end_obj_idx + 1]
@@ -111,7 +76,7 @@ def df_read_spec(df, row_count, row_start_idx, row_end_idx):
 class BaseObjectStorageDataset():        
     def __init__(self, 
                  glob, 
-                 header = None, 
+                 header = 'infer', 
                  index_metadata = None,
                  metadata_glob = None, 
                  make_metadata_df_fn = make_metadata_df, 
@@ -122,7 +87,6 @@ class BaseObjectStorageDataset():
                  max_shards = None,
                  storage_options = None, 
                  cache_dir = None,):
-        
         assert glob, "glob must be specified"
         self.glob = glob
         
@@ -131,6 +95,8 @@ class BaseObjectStorageDataset():
         if index_metadata is None:
             if metadata_glob is None:
                 index_metadata = True
+                import warnings
+                warnings.warn("metadata_glob is not specified at initialization, attempting to proceed with index_metadata=True which can take a while")
             else:
                 index_metadata = False
         else:
@@ -192,7 +158,7 @@ class BaseObjectStorageDataset():
             
         if index_metadata is None or index_metadata is True:
             objs = [f"{protocol}://{obj}" for obj in self.objs]
-            self.metadata_df = make_metadata_df_fn(df = objs_to_metadata_df_fn(objs))
+            self.metadata_df = make_metadata_df_fn(df = objs_to_metadata_df_fn(objs, header = self.header))
         else:            
             self.metadata_df = make_metadata_df_fn(glob = metadata_glob)
             
@@ -207,30 +173,50 @@ class BaseObjectStorageDataset():
         while self.max_shards:
             try:
                 num_rows, row_start_idx, row_end_idx = self.shard_size, shard_start_idx, shard_start_idx + self.shard_size
-                
-                #query rows
-                row_start_idx = row_start_idx % self.row_count
-                row_end_idx = row_end_idx % self.row_count
 
-                if self.header is None:
-                    src_df_list = [pd.read_csv(f"filecache::{uri}", header = None, nrows=nrows, skiprows=skiprows) for
-                     (skiprows, nrows, uri) in df_read_spec(self.metadata_df, self.row_count, row_start_idx, row_end_idx)]
-                elif self.header == 'infer' or self.header == 0:
-                    src_df_list = []
-                    for (skiprows, nrows, uri) in df_read_spec(self.metadata_df, self.row_count, row_start_idx, row_end_idx):
-                        if (skiprows == 0):
-                            src_df_list.append(pd.read_csv(f"filecache::{uri}", header=None, nrows = nrows, skiprows = 1))
-                        else:
-                            src_df_list.append(pd.read_csv(f"filecache::{uri}", header=None, nrows = nrows, skiprows = skiprows))
-                else:
-                    raise ValueError(f"header attribute cannot be set to {self.header}")
+                # the for loop handles the case where the shard
+                # is larger than the number of rows in the dataset
+                row_start_indicies = list(range(row_start_idx, row_end_idx, self.row_count))
+                row_end_indicies = list(row_start_indicies[1:] + [row_end_idx])
+                src_df_list = []
+                for row_start_idx, row_end_idx in zip(row_start_indicies, row_end_indicies):
 
-                # map(lambda df: df.columns, src_df_list)
-                # assert len(set(map(lambda df: df.shape[1], src_df_list))) == 1, f"Failed to process in the following fragment of the dataset {df_read_spec(self.metadata_df, self.row_count, row_start_idx, row_end_idx)} due to a change in the number of the columns."
+                    #query rows based on actual indicies in the dataset
+                    row_start_idx = row_start_idx % self.row_count
+                    row_end_idx = row_end_idx % self.row_count
+
+                    if self.header is None:
+                        for (skiprows, nrows, uri) in df_read_spec(self.metadata_df, self.row_count, row_start_idx, row_end_idx):
+                            src_df_list.append( pd.read_csv(f"filecache::{uri}", header = None, nrows=nrows, skiprows=skiprows) )
+
+                    elif self.header == 'infer' or self.header == 0:                        
+                        for (skiprows, nrows, uri) in df_read_spec(self.metadata_df, self.row_count, row_start_idx, row_end_idx):
+                            
+                            if (skiprows == 0):
+                                #skiprows == 0 when we are at the top of the file with a (potential) header                            
+                                df = pd.read_csv(f"filecache::{uri}", header = self.header, nrows = nrows, skiprows = 0)
+                                
+                                #silly hack because pandas doesn't let me to reset the index along the columns axis
+                                df = df.T
+                                df.reset_index(drop = True, inplace = True)
+                                df = df.T
+
+                            else:
+                                #(skiprows+1) skip the extra row  to account for the skipped header
+                                df = pd.read_csv(f"filecache::{uri}", header = None, nrows = nrows, skiprows = skiprows + 1) 
+
+                            src_df_list.append(df)
+
+                    else:
+                        raise ValueError(f"header attribute cannot be set to {self.header}")
+
+                    # map(lambda df: df.columns, src_df_list)
+                    # assert len(set(map(lambda df: df.shape[1], src_df_list))) == 1, f"Failed to process in the following fragment of the dataset {df_read_spec(self.metadata_df, self.row_count, row_start_idx, row_end_idx)} due to a change in the number of the columns."
 
                 yield pd.concat(
                     objs = src_df_list,
-                    copy = False
+                    copy = False,
+                    ignore_index=True,
                 )
                                 
             finally:
